#include "AGILe/CCPythia/CCPythia.hh"


namespace AGILe {

  /// @author Lars Sonnenschein
  /// @author Andy Buckley


  /// Standard constructor
  CCPythia::CCPythia() {

    // Set up particle names map
    /// @todo Also deal with gamma/e- etc. for resolved photoproduction...
    _particleNames[ELECTRON]    = "e-";
    _particleNames[POSITRON]    = "e+";
    _particleNames[PROTON]      = "p+";
    _particleNames[ANTIPROTON]  = "pbar-";
    _particleNames[NEUTRON]     = "n0";
    _particleNames[ANTINEUTRON] = "nbar0";
    _particleNames[PHOTON]      = "gamma"; 
    _particleNames[MUON]        = "mu-";
    _particleNames[ANTIMUON]    = "mu+";
    // TAU, ANTITAU
    _particleNames[NU_E]        = "nu_e";
    _particleNames[NU_EBAR]     = "nu_ebar";
    _particleNames[NU_MU]       = "nu_mu";
    _particleNames[NU_MUBAR]    = "nu_mubar";
    _particleNames[NU_TAU]      = "nu_tau";
    _particleNames[NU_TAUBAR]   = "nu_taubar";
    _particleNames[PIPLUS]      = "pi+";
    _particleNames[PIMINUS]     = "pi-";
    // PIZERO
    // PHOTOELECTRON, PHOTOPOSITRON,
    // PHOTOMUON,     PHOTOANTIMUON,
    // PHOTOTAU,      PHOTOANTITAU,
    /// For Pythia, other "normal" particles are:
    /// * gamma/e-, gamma/e+, gamma/mu-, gamma/mu+, gamma/tau-, gamma/tau+, pi0
    /// Also, with crappy PDF parameterisations and other defects:
    /// * K+, K-, KS0, KL0, Lambda0, Sigma-, Sigma0, Sigma+, Xi-, Xi0, Omega
    /// And, not useable at the moment:
    /// * pomeron, reggeon
  }
  
  
  void CCPythia::setGenSpecificInitialState(PdgCode p1, double e1, 
                                            PdgCode p2, double e2) {
    getLog() << Log::INFO << "Setting initial state..." << endl;
    // For Pythia, we have to hold on to this state until initialize() is called.
    _particleName1 = _particleNames[p1];
    _particleName2 = _particleNames[p2];
    _p1 = p1;
    _p2 = p2;
    /// @todo Get momenta and energies unconfused.
    _cmsMomentum = e1 + e2;
  }



  /// Set up default params etc.
  void CCPythia::initialize() {
    Generator::initialize();
    getLog() << Log::INFO << "Initialising..." << endl;
    _pythia.init(_p1, _p2, _cmsMomentum);
    _initialstateset = true;
    _initialized = true;
  }


  // Set random number seed
  void CCPythia::setSeed(const int seed) {
    Generator::setSeed(seed);
    _pythia.readString("Random:setSeed = on");
    Pythia8::Rndm::init(seed);
  }

  
  // Set string parameters
  bool CCPythia::setParam(const string& name, const string& value) {
    Generator::setParam(name, value);
    _pythia.readString(name + " = " + value);
    return SUCCESS;
  }
  

  // Run the generator for one event
  void CCPythia::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    int tryCount = 0;
    while (!_pythia.next() && tryCount != 10) {
      ++tryCount;
      getLog() << Log::WARN << "Event aborted - trying again" << endl;
    }
    if (tryCount == 10) {
      getLog() << Log::WARN << "Event generation aborted prematurely" << endl;
      throw runtime_error("Pythia 8 event generation aborted prematurely");
    }
    CCPythia::fillEvent(evt);
  }


  /// Fill a HepMC event
  void CCPythia::fillEvent(HepMC::GenEvent& evt) {
    _toHepMC.fill_next_event(_pythia.event, &evt);
    //getLog() << Log::INFO << "HepMC event record filling failed!" << endl;
    //throw runtime_error("HepMC event record filling failed!");

  }

  /// Get the cross section as determined by Pythia8
  const double CCPythia::getCrossSection() {
    // Return cross section in pb
    return _pythia.info.sigmaGen() * 1e09;
  }

  /// Tidy up, print out run stats, etc.
  void CCPythia::finalize() {
    getLog() << Log::INFO << "Finalising..." << endl;
    _pythia.statistics();
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::CCPythia(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
