// -*- C++ -*-
#include "AGILe/FHerwig/FHerwig.hh"
#include "AGILe/FHerwig/FHerwigWrapper65.hh"
#include "AGILe/Utils.hh"


namespace AGILe {

  // Standard constructor
  FHerwig::FHerwig() {
    /// Herwig 6.5 uses HEPEVT with 4000/10000 entries and 8-byte floats
    HepMC::HEPEVT_Wrapper::set_max_number_entries(FC_HWGETHEPEVTSIZE());
    HepMC::HEPEVT_Wrapper::set_sizeof_real(8);

    // Set up particle names map: note that they must be 8 characters long
    _particleNames[ELECTRON]    = "E-      ";
    _particleNames[POSITRON]    = "E+      ";
    _particleNames[PROTON]      = "P       ";
    _particleNames[ANTIPROTON]  = "PBAR    ";
    _particleNames[NEUTRON]     = "N       ";
    _particleNames[ANTINEUTRON] = "NBAR    ";
    _particleNames[PHOTON]      = "GAMMA   ";
    _particleNames[MUON]        = "MU-     ";
    _particleNames[ANTIMUON]    = "MU+     ";
    // TAU, ANTITAU
    _particleNames[NU_E]        = "NU_E    ";
    _particleNames[NU_EBAR]     = "NU_EBAR ";
    _particleNames[NU_MU]       = "NU_MU   ";
    _particleNames[NU_MUBAR]    = "NU_MUBAR";
    _particleNames[NU_TAU]      = "NU_TAU  ";
    _particleNames[NU_TAUBAR]   = "NU_TAUBR";
    _particleNames[PIPLUS]      = "PI+     ";
    _particleNames[PIMINUS]     = "PI-     ";
    // PIZERO
    // PHOTOELECTRON, PHOTOPOSITRON,
    // PHOTOMUON,     PHOTOANTIMUON,
    // PHOTOTAU,      PHOTOANTITAU,

    // Initialize and set default parameters (no echoing)
    FC_HWIGIN(); // Initialise common blocks
    //FC_HWPROC.IPROC = 1706; // Use qq -> ttbar production as the default process
    FC_HWPROC.IPROC = 1500; // Use QCD 2->2 as the default process
    FC_HWPROC.MAXEV = 100000000; // Maximum number of events (irrelevant!)
    FC_HWEVNT.MAXER = 100000000; // Maximum number of allowed errors
    FC_HWPRAM.PRNDEF = 0; // Enable/disable ASCII output
    FC_HWEVNT.MAXPR = 0; // Number of events to print
    FC_HWPRAM.IPRINT = 1; // Type of info to print
  }


  // Set up initial state from supplied params
  void FHerwig::setGenSpecificInitialState(int p1, double e1, int p2, double e2) {
    getLog() << Log::INFO << "Setting initial state..." << endl;
    // Herwig's initial state particles specification must be 8 chars long
    for ( unsigned int i = 0; i < 8; ++i ) {
      FC_HWBMCH.PART1[i] = _particleNames[ParticleName(p1)].c_str()[i];
      FC_HWBMCH.PART2[i] = _particleNames[ParticleName(p2)].c_str()[i];
    }
    // Set momenta / energies
    setParam("PBEAM1", e1); /// @todo Get momenta and energies unconfused?
    setParam("PBEAM2", e2); /// @todo Store a list of particle masses for this?
    /// @todo Beam ordering is significant (segfault if "wrong") for HERA initial state
  }


  // Run generator initialization
  void FHerwig::initialize() {
    Generator::initialize();
    // Compute parameter-dependent constants
    FC_HWUINC(); 
    // Initialise elementary process
    FC_HWEINI();
    _initialized = true;
  }


  // Set random number seed
  void FHerwig::setSeed(const int value) {
    Generator::setSeed(value);
    FC_HWEVNT.NRN[0] = value;
    const int nextSeed = value + SEED_OFFSET;
    FC_HWEVNT.NRN[1] = nextSeed;
    Generator::setSeed(nextSeed);
  }


  // Set parameters.
  bool FHerwig::setParam(const string& name, const string& value) {
    Generator::setParam(name, value);

    // Strings
    if (name == "AUTPDF(1)" || name == "AUTPDF") { 
      getLog() << Log::INFO << "Setting AUTPDF(1) = " << value << endl;
      int nch = (value.size() < AUTPDF_LENGTH) ? value.size() : AUTPDF_LENGTH-1;
      for(int i = 0; i < AUTPDF_LENGTH; i++) {
        if(i < nch) {
          FC_HWPRCH.AUTPDF[0][i] = value.data()[i];
        } else {
          FC_HWPRCH.AUTPDF[0][i] = ' ';
        }
      }
    }
    if (name == "AUTPDF(2)" || name == "AUTPDF") { 
      getLog() << Log::INFO << "Setting AUTPDF(2) = " << value << endl;
      int nch = (value.size() < AUTPDF_LENGTH) ? value.size() : AUTPDF_LENGTH-1;
      for(int i = 0; i < AUTPDF_LENGTH; i++) {
        if(i < nch) {
          FC_HWPRCH.AUTPDF[1][i] = value.data()[i];
        } else {
          FC_HWPRCH.AUTPDF[1][i] = ' ';
        }
      }
    }

    // Integers
    if (name == "IPROC") {  
      // 1610 = gg -> H -> WW; 1706 = qq -> ttbar; 2510 = ttH -> ttWW, ...
      getLog() << Log::INFO << "Setting process code (IPROC) = " << asInt(value) << endl;
      FC_HWPROC.IPROC = asInt(value);
      //} else if (name == "MAXEV") { 
      //getLog() << Log::INFO << "Setting maximum number of events (MAXEV) = " << asInt(value) << endl;
      //FC_HWPROC.MAXEV = asInt(value);
    } else if (name == "MAXPR") {
      getLog() << Log::INFO << "Setting max event printouts (MAXPR) = " << asInt(value) << endl;
      FC_HWEVNT.MAXPR = asInt(value);
    } else if (name == "IPRINT") { 
      getLog() << Log::INFO << "Setting printout info code (IPRINT) = " << asInt(value) << endl;
      FC_HWPRAM.IPRINT = asInt(value);
    } else if (name == "CLDIR1") { 
      getLog() << Log::INFO << "Setting ... (CLDIR(1)) = " << asInt(value) << endl;
      FC_HWPRAM.CLDIR[0] = asInt(value);
    } else if (name == "CLDIR2") { 
      getLog() << Log::INFO << "Setting ... (CLDIR(2)) = " << asInt(value) << endl;
      FC_HWPRAM.CLDIR[1] = asInt(value);
    } else if (name == "IOPREM") { 
      getLog() << Log::INFO << "Setting ... (IOPREM) = " << asInt(value) << endl;
      FC_HWPRAM.IOPREM = asInt(value);
    } else if (name == "ISPAC") { 
      getLog() << Log::INFO << "Setting ... (ISPAC) = " << asInt(value) << endl;
      FC_HWPRAM.ISPAC = asInt(value);
    } else if (name == "NFLAV") { 
      getLog() << Log::INFO << "Setting ... (NFLAV) = " << asInt(value) << endl;
      FC_HWPRAM.NFLAV = asInt(value);
    } else if (name == "NSTRU") { 
      getLog() << Log::INFO << "Setting ... (NSTRU) = " << asInt(value) << endl;
      FC_HWPRAM.NSTRU = asInt(value);
    } else if (name == "MODPDF") { 
      getLog() << Log::INFO << "Setting MODPDF(1&2) = " << asInt(value) << endl;
      FC_HWPRAM.MODPDF[0] = asInt(value);
      FC_HWPRAM.MODPDF[1] = asInt(value);
    } else if (name == "MODPDF(1)") { 
      getLog() << Log::INFO << "Setting MODPDF(1) = " << asInt(value) << endl;
      FC_HWPRAM.MODPDF[0] = asInt(value);
    } else if (name == "MODPDF(2)") { 
      getLog() << Log::INFO << "Setting MODPDF(2) = " << asInt(value) << endl;
      FC_HWPRAM.MODPDF[1] = asInt(value);
    } else if (name == "IFLMAX") { 
      getLog() << Log::INFO << "Setting IFLMAX = " << asInt(value) << endl;
      FC_HWHARD.IFLMAX = asInt(value);
      //} else if (name == "MAXER") { 
      //getLog() << Log::INFO << "Setting MAXER = " << asInt(value) << endl;
      //FC_HWEVNT.MAXER = asInt(value);
    } //...

    // Doubles
    else if (name == "PBEAM1") { 
      getLog() << Log::INFO << "Setting beam 1 momentum (PBEAM1) = " << asDouble(value) << endl;
      FC_HWPROC.PBEAM1 = asDouble(value);
    } else if (name == "PBEAM2") { 
      getLog() << Log::INFO << "Setting beam 2 momentum (PBEAM2) = " << asDouble(value) << endl;
      FC_HWPROC.PBEAM2 = asDouble(value);
    } else if (name == "EBEAM1") { 
      getLog() << Log::INFO << "Setting beam 1 energy (EBEAM1) = " << asDouble(value) << endl;
      FC_HWPROC.EBEAM1 = asDouble(value);
    } else if (name == "EBEAM2") { 
      getLog() << Log::INFO << "Setting beam 2 energy (EBEAM2) = " << asDouble(value) << endl;
      FC_HWPROC.EBEAM2 = asDouble(value);
    } else if (name == "PTMIN") { 
      getLog() << Log::INFO << "Setting minimum hard pt (PTMIN) = " << asDouble(value) << endl;
      FC_HWHARD.PTMIN = asDouble(value); 
    } else if (name == "PTMAX") { 
      getLog() << Log::INFO << "Setting maximum hard pt (PTMAX) = " << asDouble(value) << endl;
      FC_HWHARD.PTMAX = asDouble(value); 
    } else if (name == "Q2WWMN") { 
      getLog() << Log::INFO << "Setting .... (Q2WWMN) = " << asDouble(value) << endl;
      FC_HWHARD.Q2WWMN = asDouble(value); 
    } else if (name == "Q2WWMX") { 
      getLog() << Log::INFO << "Setting .... (Q2WWMX) = " << asDouble(value) << endl;
      FC_HWHARD.Q2WWMX = asDouble(value); 
    } else if (name == "YWWMIN") { 
      getLog() << Log::INFO << "Setting ... (YWWMIN) = " << asDouble(value) << endl;
      FC_HWHARD.YWWMIN = asDouble(value); 
    } else if (name == "YWWMAX") { 
      getLog() << Log::INFO << "Setting ... (YWWMAX) = " << asDouble(value) << endl;
      FC_HWHARD.YWWMAX = asDouble(value); 
    } else if ( name =="EMMIN") {
      getLog() << Log::INFO << "Setting ... (EMMIN) = "<< asDouble(value) << endl;
      FC_HWHARD.EMMIN = asDouble(value);
    } else if (name == "PRECO") { 
      getLog() << Log::INFO << "Setting ... (PRECO) = " << asDouble(value) << endl;
      FC_HWUCLU.PRECO = asDouble(value); 
    } else if (name == "PDFX0") { 
      getLog() << Log::INFO << "Setting maximum x for saturation effects (PDFX0) = " << asDouble(value) << endl;
      FC_HW6506.PDFX0 = asDouble(value); 
    } else if (name == "PDFPOW") { 
      getLog() << Log::INFO << "Setting suppression of PDFs below x0 (PDFPOW) = " << asDouble(value) << endl;
      FC_HW6506.PDFPOW = asDouble(value); 
    } else if (name == "BTCLM") { 
      getLog() << Log::INFO << "Setting ... (BTCLM) = " << asDouble(value) << endl;
      FC_HWPRAM.BTCLM = asDouble(value);
    } else if (name == "CLMAX") { 
      getLog() << Log::INFO << "Setting ... (CLMAX) = " << asDouble(value) << endl;
      FC_HWPRAM.CLMAX = asDouble(value);
    } else if (name == "CLPOW") { 
      getLog() << Log::INFO << "Setting ... (CLPOW) = " << asDouble(value) << endl;
      FC_HWPRAM.CLPOW = asDouble(value);
    } else if (name == "CLSMR1") { 
      getLog() << Log::INFO << "Setting ... (CLSMR(1)) = " << asDouble(value) << endl;
      FC_HWPRAM.CLSMR[0] = asDouble(value);
    } else if (name == "CLSMR2") { 
      getLog() << Log::INFO << "Setting ... (CLSMR(2)) = " << asDouble(value) << endl;
      FC_HWPRAM.CLSMR[1] = asDouble(value);
    } else if (name == "PRSOF") { 
      getLog() << Log::INFO << "Setting ... (PRSOF) = " << asDouble(value) << endl;
      FC_HWPRAM.PRSOF = asDouble(value);
    } else if (name == "PSPLT1") { 
      getLog() << Log::INFO << "Setting ... (PSPLT(1)) = " << asDouble(value) << endl;
      FC_HWPRAM.PSPLT[0] = asDouble(value);
    } else if (name == "PSPLT2") { 
      getLog() << Log::INFO << "Setting ... (PSPLT(2)) = " << asDouble(value) << endl;
      FC_HWPRAM.PSPLT[1] = asDouble(value);
    } else if (name == "PTRMS") { 
      getLog() << Log::INFO << "Setting ... (PTRMS) = " << asDouble(value) << endl;
      FC_HWPRAM.PTRMS = asDouble(value);
    } else if (name == "QCDLAM") { 
      getLog() << Log::INFO << "Setting ... (QCDLAM) = " << asDouble(value) << endl;
      FC_HWPRAM.QCDLAM = asDouble(value);
    } else if (name == "QSPAC") { 
      getLog() << Log::INFO << "Setting ... (QSPAC) = " << asDouble(value) << endl;
      FC_HWPRAM.QSPAC = asDouble(value);
    } else if (name == "VGCUT") { 
      getLog() << Log::INFO << "Setting ... (VGCUT) = " << asDouble(value) << endl;
      FC_HWPRAM.VGCUT = asDouble(value);
    } else if (name == "VQCUT") { 
      getLog() << Log::INFO << "Setting ... (VQCUT) = " << asDouble(value) << endl;
      FC_HWPRAM.VQCUT = asDouble(value);
    } else if (name == "VPCUT") { 
      getLog() << Log::INFO << "Setting ... (VPCUT) = " << asDouble(value) << endl;
      FC_HWPRAM.VPCUT = asDouble(value);
    } else if (name == "OMEGA0") { 
      getLog() << Log::INFO << "Setting ... (OMEGA0) = " << asDouble(value) << endl;
      FC_HWHARD.OMEGA0 = asDouble(value);
    } else if (name == "ASFIXD") { 
      getLog() << Log::INFO << "Setting ... (ASFIXD) = " << asDouble(value) << endl;
      FC_HWHARD.ASFIXD = asDouble(value);
    } else if (name == "Q2MIN") {
      getLog() << Log::INFO << "Setting ... (Q2MIN) = " << asDouble(value) << endl;
      FC_HWHARD.Q2MIN = asDouble(value);
    } else if (name == "Q2MAX") {
      getLog() << Log::INFO << "Setting ... (Q2MAX) = " << asDouble(value) << endl;
      FC_HWHARD.Q2MAX = asDouble(value);
    } else if (name == "YBMIN") {
      getLog() << Log::INFO << "Setting ... (YBMIN) = " << asDouble(value) << endl;
      FC_HWHARD.YBMIN = asDouble(value);
    } else if (name == "YBMAX") {
      getLog() << Log::INFO << "Setting ... (YBMAX) = " << asDouble(value) << endl;
      FC_HWHARD.YBMAX = asDouble(value);
    } //...

    // Booleans (done as ints in Fortran)
    else if (name == "GENSOF") { 
      getLog() << Log::INFO << "Setting ... (GENSOF) = " << asInt(value) << endl;
      FC_HWEVNT.GENSOF = asInt(value);
    } else if (name == "AZSOFT") { 
      getLog() << Log::INFO << "Setting ... (AZSOFT) = " << asInt(value) << endl;
      FC_HWPRAM.AZSOFT = asInt(value);
    } else if (name == "CLRECO") { 
      getLog() << Log::INFO << "Setting ... (CLRECO) = " << asInt(value) << endl;
      FC_HWUCLU.CLRECO = asInt(value);
    } else if (name == "AZSPIN") { 
      getLog() << Log::INFO << "Setting ... (AZSPIN) = " << asInt(value) << endl;
      FC_HWPRAM.AZSPIN = asInt(value);
    } else if (name == "ZPRIME") { 
      getLog() << Log::INFO << "Setting Z' flag (ZPRIME) = " << asInt(value) << endl;
      FC_HWPRAM.ZPRIME = asInt(value);
    } else if (name == "HARDME") { 
      getLog() << Log::INFO << "Setting hard matrix element flag (HARDME) = " << asInt(value) << endl;
      FC_HWPRAM.HARDME = asInt(value);
    } else if (name == "SOFTME") { 
      getLog() << Log::INFO << "Setting soft matrix element flag (SOFTME) = " << asInt(value) << endl;
      FC_HWPRAM.SOFTME = asInt(value);
    } else if (name == "NOSPAC") { 
      getLog() << Log::INFO << "Setting ... (NOSPAC) = " << asInt(value) << endl;
      FC_HWPRAM.NOSPAC = asInt(value);
    } else if (name == "PRVTX") { 
      getLog() << Log::INFO << "Setting ... (PRVTX) = " << asInt(value) << endl;
      FC_HWPRAM.PRVTX = asInt(value);
    } else if (name == "PRNDEC") { 
      getLog() << Log::INFO << "Setting ... (PRNDEC) = " << asInt(value) << endl;
      FC_HWPRAM.PRNDEC = asInt(value);
    } else if (name == "PRNDEF") { 
      getLog() << Log::INFO << "Setting ... (PRNDEF) = " << asInt(value) << endl;
      FC_HWPRAM.PRNDEF = asInt(value);
    } else if (name == "PRNTEX") { 
      getLog() << Log::INFO << "Setting LaTeX output flag (PRNTEX) = " << asInt(value) << endl;
      FC_HWPRAM.PRNTEX = asInt(value);
    } else if (name == "PRNWEB") { 
      getLog() << Log::INFO << "Setting HTML output flag (PRNWEB) = " << asInt(value) << endl;
      FC_HWPRAM.PRNWEB = asInt(value);
    } //...

    // Error
    else {
      getLog() << Log::ERROR << "FHerwig doesn't have a parameter called " << name << endl;
      return FAILURE;
    }
    return SUCCESS;
  }


  // Run the generator for one event
  void FHerwig::makeEvent(HepMC::GenEvent& evt) {
    // Loop until event works
    while (true) {
      Generator::makeEvent(evt);
      FC_HWUINE();  // Initialize event
      FC_HWEPRO();  // Generate hard subprocess
      FC_HWBGEN();  // Generate parton cascade
      FC_HWDHOB();  // Do heavy quark decays
      FC_HWCFOR();  // Do cluster formation
      FC_HWCDEC();  // Do cluster decays
      FC_HWDHAD();  // Do unstable particle decays
      FC_HWDHVY();  // Do heavy flavor decays
      FC_HWMEVT();  // Add soft underlying event
      FC_HWUFNE();  // Finish event
      if (FC_HWEVNT.IERROR == 0) break;
    }
    fillEvent(evt);
  }


  /// Fill a HepMC event
  void FHerwig::fillEvent(HepMC::GenEvent& evt) {
    HepMC::IO_HERWIG hepevt;
    hepevt.fill_next_event(&evt);
  }
  

  /// Return the beam number (0 or 1) of a particle.  
  /// Return -1 if not a beam particle
  int FHerwig::beamNumber(PdgCode pid){
    string pName = _particleNames[pid];
    string beamName = "";
    for (int i = 0; i < 8; ++i) {
      beamName += FC_HWBMCH.PART1[i];
    }
    if (beamName.compare(pName) == 0) return 0;
    
    for (int i = 0; i < 8; ++i) {
      beamName += FC_HWBMCH.PART2[i];
    }
    if (beamName.compare(pName) == 0) return 1;    
    return -1;
  }
  

  string FHerwig::getPDFSet(PdgCode pid){    
    const int beamNum = beamNumber(pid);
    if (beamNum < 0) {
      throw runtime_error("PDFSet unknown for PDG code " + pid);
    }
    string autString = "";
    for (int ii = 0; ii != AUTPDF_LENGTH; ++ii){
      autString += FC_HWPRCH.AUTPDF[beamNum][ii];
    }
    return autString;
  }

  
  int FHerwig::getPDFMember(PdgCode pid){
    const int beamNum = beamNumber(pid);    
    if (beamNum < 0) {
      throw runtime_error("PDFMember unknown for PDG code "+ pid);
    }
    const int member = FC_HWPRAM.MODPDF[beamNum];
    return member;
  }
  

  string FHerwig::getPDFScheme(PdgCode pid)const{
    return "LHAPDF";
  }

  
  const double FHerwig::getCrossSection(){
    _crossSection = FC_HWEVNT.AVWGT * 1000.0;
    return _crossSection;
  }
  
  
  /// Tidy up, print out run stats, etc.
  void FHerwig::finalize() {
    getLog() << Log::INFO << "Finalising..." << endl;
    FC_HWEFIN();
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::FHerwig(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
